@@ -1,5 +1,7 @@
 import asyncio
+import signal
 import ssl
+import threading
 
 try:
     import aiohttp
@@ -12,6 +14,24 @@ from . import exceptions
 from . import packet
 from . import payload
 
+async_signal_handler_set = False
+
+
+def async_signal_handler():
+    """SIGINT handler.
+
+    Disconnect all active async clients.
+    """
+    async def _handler():
+        asyncio.get_event_loop().stop()
+        for c in client.connected_clients[:]:
+            if c.is_asyncio_based():
+                await c.disconnect()
+        else:  # pragma: no cover
+            pass
+
+    asyncio.ensure_future(_handler())
+
 
 class AsyncClient(client.Client):
     """An Engine.IO client for asyncio.
@@ -30,6 +50,10 @@ class AsyncClient(client.Client):
                  versions.
     :param request_timeout: A timeout in seconds for requests. The default is
                             5 seconds.
+    :param http_session: an initialized ``aiohttp.ClientSession`` object to be
+                         used when sending requests to the server. Use it if
+                         you need to add special client options such as proxy
+                         servers, SSL certificates, etc.
     :param ssl_verify: ``True`` to verify SSL certificates, or ``False`` to
                        skip SSL certificate verification, allowing
                        connections to servers with self signed certificates.
@@ -38,7 +62,7 @@ class AsyncClient(client.Client):
     def is_asyncio_based(self):
         return True
 
-    async def connect(self, url, headers={}, transports=None,
+    async def connect(self, url, headers=None, transports=None,
                       engineio_path='engine.io'):
         """Connect to an Engine.IO server.
 
@@ -61,6 +85,17 @@ class AsyncClient(client.Client):
             eio = engineio.Client()
             await eio.connect('http://localhost:5000')
         """
+        global async_signal_handler_set
+        if not async_signal_handler_set and \
+                threading.current_thread() == threading.main_thread():
+
+            try:
+                asyncio.get_event_loop().add_signal_handler(
+                    signal.SIGINT, async_signal_handler)
+                async_signal_handler_set = True
+            except NotImplementedError:  # pragma: no cover
+                self.logger.warning('Signal handler is unsupported')
+
         if self.state != 'disconnected':
             raise ValueError('Client is not in a disconnected state')
         valid_transports = ['polling', 'websocket']
@@ -74,7 +109,7 @@ class AsyncClient(client.Client):
         self.transports = transports or valid_transports
         self.queue = self.create_queue()
         return await getattr(self, '_connect_' + self.transports[0])(
-            url, headers, engineio_path)
+            url, headers or {}, engineio_path)
 
     async def wait(self):
         """Wait until the connection with the server ends.
@@ -183,9 +218,14 @@ class AsyncClient(client.Client):
             raise exceptions.ConnectionError(
                 'Connection refused by the server')
         if r.status < 200 or r.status >= 300:
+            self._reset()
+            try:
+                arg = await r.json()
+            except aiohttp.ClientError:
+                arg = None
             raise exceptions.ConnectionError(
                 'Unexpected status code {} in server response'.format(
-                    r.status))
+                    r.status), arg)
         try:
             p = payload.Payload(encoded_payload=await r.read())
         except ValueError:
@@ -199,8 +239,8 @@ class AsyncClient(client.Client):
             'Polling connection accepted with ' + str(open_packet.data))
         self.sid = open_packet.data['sid']
         self.upgrades = open_packet.data['upgrades']
-        self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-        self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+        self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+        self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
         self.current_transport = 'polling'
         self.base_url += '&sid=' + self.sid
 
@@ -249,7 +289,7 @@ class AsyncClient(client.Client):
         for header, value in headers.items():
             if header.lower() == 'cookie':
                 cookies = dict(
-                    [cookie.split('=') for cookie in value.split('; ')])
+                    [cookie.split('=', 1) for cookie in value.split('; ')])
                 del headers[header]
                 break
         self.http.cookie_jar.update_cookies(cookies)
@@ -319,8 +359,8 @@ class AsyncClient(client.Client):
                 'WebSocket connection accepted with ' + str(open_packet.data))
             self.sid = open_packet.data['sid']
             self.upgrades = open_packet.data['upgrades']
-            self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-            self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+            self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+            self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
             self.current_transport = 'websocket'
 
             self.state = 'connected'
@@ -506,12 +546,19 @@ class AsyncClient(client.Client):
                 break
             except Exception as e:
                 self.logger.info(
-                    'Unexpected error "%s", aborting', str(e))
+                    'Unexpected error receiving packet: "%s", aborting',
+                    str(e))
                 await self.queue.put(None)
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
-            pkt = packet.Packet(encoded_packet=p)
+            try:
+                pkt = packet.Packet(encoded_packet=p)
+            except Exception as e:  # pragma: no cover
+                self.logger.info(
+                    'Unexpected error decoding packet: "%s", aborting', str(e))
+                await self.queue.put(None)
+                break
             await self._receive_packet(pkt)
 
         self.logger.info('Waiting for write loop task to end')
@@ -589,7 +636,8 @@ class AsyncClient(client.Client):
                             await self.ws.send_str(pkt.encode(
                                 always_bytes=False))
                         self.queue.task_done()
-                except aiohttp.client_exceptions.ServerDisconnectedError:
+                except (aiohttp.client_exceptions.ServerDisconnectedError,
+                        BrokenPipeError, OSError):
                     self.logger.info(
                         'Write loop: WebSocket connection was closed, '
                         'aborting')
