@@ -1,3 +1,5 @@
+from base64 import b64encode
+from json import JSONDecodeError
 import logging
 try:
     import queue
@@ -35,10 +37,8 @@ def signal_handler(sig, frame):
     Disconnect all active clients and then invoke the original signal handler.
     """
     for client in connected_clients[:]:
-        if client.is_asyncio_based():
-            client.start_background_task(client.disconnect, abort=True)
-        else:
-            client.disconnect(abort=True)
+        if not client.is_asyncio_based():
+            client.disconnect()
     if callable(original_signal_handler):
         return original_signal_handler(sig, frame)
     else:  # pragma: no cover
@@ -46,7 +46,7 @@ def signal_handler(sig, frame):
         return signal.default_int_handler(sig, frame)
 
 
-original_signal_handler = signal.signal(signal.SIGINT, signal_handler)
+original_signal_handler = None
 
 
 class Client(object):
@@ -57,13 +57,18 @@ class Client(object):
 
     :param logger: To enable logging set to ``True`` or pass a logger object to
                    use. To disable logging set to ``False``. The default is
-                   ``False``.
+                   ``False``. Note that fatal errors are logged even when
+                   ``logger`` is ``False``.
     :param json: An alternative json module to use for encoding and decoding
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
                  versions.
     :param request_timeout: A timeout in seconds for requests. The default is
                             5 seconds.
+    :param http_session: an initialized ``requests.Session`` object to be used
+                         when sending requests to the server. Use it if you
+                         need to add special client options such as proxy
+                         servers, SSL certificates, etc.
     :param ssl_verify: ``True`` to verify SSL certificates, or ``False`` to
                        skip SSL certificate verification, allowing
                        connections to servers with self signed certificates.
@@ -75,7 +80,16 @@ class Client(object):
                  logger=False,
                  json=None,
                  request_timeout=5,
+                 http_session=None,
                  ssl_verify=True):
+        global original_signal_handler
+        if original_signal_handler is None and \
+                threading.current_thread() == threading.main_thread():
+            if not hasattr(signal, 'SIGBREAK'):
+                sig = signal.SIGINT
+            else:  # pragma: no cover
+                sig = signal.SIGBREAK
+            original_signal_handler = signal.signal(sig, signal_handler)
         self.handlers = {}
         self.base_url = None
         self.transports = None
@@ -85,7 +99,7 @@ class Client(object):
         self.ping_interval = None
         self.ping_timeout = None
         self.pong_received = True
-        self.http = None
+        self.http = http_session
         self.ws = None
         self.read_loop_task = None
         self.write_loop_task = None
@@ -147,7 +161,7 @@ class Client(object):
             return set_handler
         set_handler(handler)
 
-    def connect(self, url, headers={}, transports=None,
+    def connect(self, url, headers=None, transports=None,
                 engineio_path='engine.io'):
         """Connect to an Engine.IO server.
 
@@ -181,7 +195,7 @@ class Client(object):
         self.transports = transports or valid_transports
         self.queue = self.create_queue()
         return getattr(self, '_connect_' + self.transports[0])(
-            url, headers, engineio_path)
+            url, headers or {}, engineio_path)
 
     def wait(self):
         """Wait until the connection with the server ends.
@@ -289,9 +303,14 @@ class Client(object):
             raise exceptions.ConnectionError(
                 'Connection refused by the server')
         if r.status_code < 200 or r.status_code >= 300:
+            self._reset()
+            try:
+                arg = r.json()
+            except JSONDecodeError:
+                arg = None
             raise exceptions.ConnectionError(
                 'Unexpected status code {} in server response'.format(
-                    r.status_code))
+                    r.status_code), arg)
         try:
             p = payload.Payload(encoded_payload=r.content)
         except ValueError:
@@ -305,8 +324,8 @@ class Client(object):
             'Polling connection accepted with ' + str(open_packet.data))
         self.sid = open_packet.data['sid']
         self.upgrades = open_packet.data['upgrades']
-        self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-        self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+        self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+        self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
         self.current_transport = 'polling'
         self.base_url += '&sid=' + self.sid
 
@@ -348,23 +367,72 @@ class Client(object):
             self.logger.info(
                 'Attempting WebSocket connection to ' + websocket_url)
 
-        # get the cookies from the long-polling connection so that they can
-        # also be sent the the WebSocket route
+        # get cookies and other settings from the long-polling connection
+        # so that they are preserved when connecting to the WebSocket route
         cookies = None
+        extra_options = {}
         if self.http:
+            # cookies
             cookies = '; '.join(["{}={}".format(cookie.name, cookie.value)
                                  for cookie in self.http.cookies])
+            for header, value in headers.items():
+                if header.lower() == 'cookie':
+                    if cookies:
+                        cookies += '; '
+                    cookies += value
+                    del headers[header]
+                    break
 
+            # auth
+            if 'Authorization' not in headers and self.http.auth is not None:
+                if not isinstance(self.http.auth, tuple):  # pragma: no cover
+                    raise ValueError('Only basic authentication is supported')
+                basic_auth = '{}:{}'.format(
+                    self.http.auth[0], self.http.auth[1]).encode('utf-8')
+                basic_auth = b64encode(basic_auth).decode('utf-8')
+                headers['Authorization'] = 'Basic ' + basic_auth
+
+            # cert
+            # this can be given as ('certfile', 'keyfile') or just 'certfile'
+            if isinstance(self.http.cert, tuple):
+                extra_options['sslopt'] = {
+                    'certfile': self.http.cert[0],
+                    'keyfile': self.http.cert[1]}
+            elif self.http.cert:
+                extra_options['sslopt'] = {'certfile': self.http.cert}
+
+            # proxies
+            if self.http.proxies:
+                proxy_url = None
+                if websocket_url.startswith('ws://'):
+                    proxy_url = self.http.proxies.get(
+                        'ws', self.http.proxies.get('http'))
+                else:  # wss://
+                    proxy_url = self.http.proxies.get(
+                        'wss', self.http.proxies.get('https'))
+                if proxy_url:
+                    parsed_url = urllib.parse.urlparse(
+                        proxy_url if '://' in proxy_url
+                        else 'scheme://' + proxy_url)
+                    print(parsed_url)
+                    extra_options['http_proxy_host'] = parsed_url.hostname
+                    extra_options['http_proxy_port'] = parsed_url.port
+                    extra_options['http_proxy_auth'] = (
+                        (parsed_url.username, parsed_url.password)
+                        if parsed_url.username or parsed_url.password
+                        else None)
+
+            # verify
+            if not self.http.verify:
+                self.ssl_verify = False
+
+        if not self.ssl_verify:
+            extra_options['sslopt'] = {"cert_reqs": ssl.CERT_NONE}
         try:
-            if not self.ssl_verify:
-                ws = websocket.create_connection(
-                    websocket_url + self._get_url_timestamp(), header=headers,
-                    cookie=cookies, sslopt={"cert_reqs": ssl.CERT_NONE})
-            else:
-                ws = websocket.create_connection(
-                    websocket_url + self._get_url_timestamp(), header=headers,
-                    cookie=cookies)
-        except (ConnectionError, IOError):
+            ws = websocket.create_connection(
+                websocket_url + self._get_url_timestamp(), header=headers,
+                cookie=cookies, enable_multithread=True, **extra_options)
+        except (ConnectionError, IOError, websocket.WebSocketException):
             if upgrade:
                 self.logger.warning(
                     'WebSocket upgrade failed: connection error')
@@ -416,8 +484,8 @@ class Client(object):
                 'WebSocket connection accepted with ' + str(open_packet.data))
             self.sid = open_packet.data['sid']
             self.upgrades = open_packet.data['upgrades']
-            self.ping_interval = open_packet.data['pingInterval'] / 1000.0
-            self.ping_timeout = open_packet.data['pingTimeout'] / 1000.0
+            self.ping_interval = int(open_packet.data['pingInterval']) / 1000.0
+            self.ping_timeout = int(open_packet.data['pingTimeout']) / 1000.0
             self.current_transport = 'websocket'
 
             self.state = 'connected'
@@ -524,7 +592,7 @@ class Client(object):
                 self.logger.info(
                     'PONG response has not been received, aborting')
                 if self.ws:
-                    self.ws.shutdown()
+                    self.ws.close(timeout=0)
                 self.queue.put(None)
                 break
             self.pong_received = False
@@ -588,12 +656,19 @@ class Client(object):
                 break
             except Exception as e:
                 self.logger.info(
-                    'Unexpected error "%s", aborting', str(e))
+                    'Unexpected error receiving packet: "%s", aborting',
+                    str(e))
                 self.queue.put(None)
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
-            pkt = packet.Packet(encoded_packet=p)
+            try:
+                pkt = packet.Packet(encoded_packet=p)
+            except Exception as e:  # pragma: no cover
+                self.logger.info(
+                    'Unexpected error decoding packet: "%s", aborting', str(e))
+                self.queue.put(None)
+                break
             self._receive_packet(pkt)
 
         self.logger.info('Waiting for write loop task to end')
@@ -669,7 +744,8 @@ class Client(object):
                         else:
                             self.ws.send(encoded_packet)
                         self.queue.task_done()
-                except websocket.WebSocketConnectionClosedException:
+                except (websocket.WebSocketConnectionClosedException,
+                        BrokenPipeError, OSError):
                     self.logger.warning(
                         'WebSocket connection was closed, aborting')
                     break
